# -*- coding: utf8 -*-
import json
import os
from collections import defaultdict

import six
from .path_utils import enumerate_paths


def __get_json_items(data, key_name=None):
    key_name = key_name or '_sha'

    if isinstance(data, dict):
        for key, val in data.items():
            val[key_name] = key

            yield val
    else:  # list, generators
        for key, val in data:
            val[key_name] = key

            yield val


def multi_line_json_from_data(data, write_to=None, key_name=None):
    result = write_to or six.BytesIO()
    for i, item in enumerate(__get_json_items(data, key_name=key_name)):
        json_str = json.dumps(item) + '\n'
        if six.PY3:
            json_str = json_str.encode()

        result.write(json_str)

    return result


def newline_json_file_from_json_file(datafile):
    json_data = json.load(datafile)
    return multi_line_json_from_data(json_data)


def json_data_from_files(data_path, files, data_per_entry):
    def rel_path_if_needed(path):
        if os.path.isabs(path):
            return os.path.relpath(path, data_path)

        return path

    for file_path in enumerate_paths(files):
        file_path = rel_path_if_needed(file_path)
        yield file_path, data_per_entry


def normalize_item(item):
    result_item = {}

    for key, val in item.items():
        if key == 'meta':
            continue

        result_item['@' + key] = val

    for key, val in item.get('meta', {}).items():
        result_item[key] = val

    return result_item


def __enum_dict_of_object(d):
    for data in d.values():
        yield data


def __enum_list_of_object(d):
    for data in d:
        yield data


def __count_dict_keys(d, enum_method):
    all_keys = defaultdict(int)
    for item in enum_method(d):
        for key in item.keys():
            all_keys[key] += 1

    return all_keys


def dict_normalize(d):
    enum_method = __enum_dict_of_object if isinstance(d, dict) else __enum_list_of_object

    all_keys = __count_dict_keys(d, enum_method)

    missing_keys = {key: total for key, total in all_keys.items() if total < len(d)}

    for item in enum_method(d):
        if len(item) == len(all_keys):
            continue

        for key in missing_keys.keys():
            if key in item:
                continue

            item[key] = None



# -*- coding: utf8 -*-


class ConnectionMixin(object):
    def __init__(self, connection):
        self.__connection = connection

        super(ConnectionMixin, self).__init__()

    @property
    def _connection(self):
        return self.__connection

    def close(self):
        self._connection.close()



# -*- coding: utf8 -*-


class Expando(object):
    pass


def build_context(http_session, output_format=None, config_prefix=None, config_file=None):
    ctx = Expando()

    init_context2(ctx, http_session, output_format, config_prefix, config_file)

    return ctx


def init_context2(ctx, http_session, output_format=None, config_prefix=None, config_file=None):
    from .config import Config
    from .api import handle_api

    ctx.obj = Expando()

    config = Config(config_prefix, config_file)

    ctx.obj.config = config
    ctx.obj.handle_api = handle_api
    ctx.obj.session = http_session

    ctx.obj.api_host = config.api_host
    ctx.obj.host = config.host
    ctx.obj.refresh_token = config.refresh_token

    ctx.obj.output_format = output_format

    ctx.obj.refresh_token = config.refresh_token
    ctx.obj.id_token = config.id_token

    ctx.obj.rm_socket_server = config.rm_socket_server
    ctx.obj.rm_config_volume = config.rm_config_volume
    ctx.obj.rm_manager_image = config.rm_manager_image
    ctx.obj.rm_container_name = config.rm_container_name



# -*- coding: utf8 -*-
from .connection_mixin import ConnectionMixin


class BackendMixin(ConnectionMixin):
    def __init__(self, connection, config, session, handle_api):
        self.__volume_id = connection.data_volume_config.volume_id
        self.__config = config
        self.__handle_api = handle_api
        self.__session = session
        super(BackendMixin, self).__init__(connection)

    @property
    def _config(self):
        return self.__config

    @property
    def _handle_api(self):
        return self.__handle_api

    @property
    def _session(self):
        return self.__session

    @property
    def _volume_id(self):
        return self.__volume_id



# -*- coding: utf8 -*-
from abc import ABCMeta, abstractmethod
from contextlib import closing

import six

from ..connection_mixin import ConnectionMixin


class BaseMLIndex(ConnectionMixin):
    __metaclass__ = ABCMeta

    def __init__(self, connection):
        super(BaseMLIndex, self).__init__(connection)

        if not self._connection.read_only:
            self._create_table_if_needed()

    @abstractmethod
    def _create_table_if_needed(self):
        """

        :return:
        """

    @abstractmethod
    def begin_commit(self, commit_sha, tree_id, ts):
        """

        :return:
        """

    @abstractmethod
    def end_commit(self):
        """

        :return:
        """

    @classmethod
    def _decode_entries(cls, entries):
        for name, entry in entries.items():
            ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags, url = entry

            if not isinstance(name, six.string_types):
                name = name.decode('utf8')

            if not isinstance(sha, six.string_types):
                sha = sha.decode('utf8')

            row = (name, sha, ctime, mtime, mode, uid, gid, size, url)

            yield row



# -*- coding: utf8 -*-
from ..scam import MLQueryVisitor


# noinspection PyClassicStyleClass
class VersionVisitor(MLQueryVisitor):
    def __init__(self):
        self.version = None

    def generic_visit(self, node, parents=None, context=None):
        pass

    def visit_function_version(self, node, parents=None, context=None):
        self.version = node.version



# file.py -- Safe access to git files
# Copyright (C) 2010 Google, Inc.
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Safe access to git files."""

import errno
import io
import os
import sys
import tempfile

import time


def ensure_dir_exists(dirname):
    """Ensure a directory exists, creating if necessary."""
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def _fancy_rename(oldname, newname):
    """Rename file with temporary backup file to rollback if rename fails"""
    if not os.path.exists(newname):
        try:
            os.rename(oldname, newname)
        except OSError:
            raise
        return

    # destination file exists
    try:
        (fd, tmpfile) = tempfile.mkstemp(".tmp", prefix=oldname+".", dir=".")
        os.close(fd)
        os.remove(tmpfile)
    except OSError:
        # either file could not be created (e.g. permission problem)
        # or could not be deleted (e.g. rude virus scanner)
        raise
    try:
        os.rename(newname, tmpfile)
    except OSError:
        raise   # no rename occurred
    try:
        os.rename(oldname, newname)
    except OSError:
        os.rename(tmpfile, newname)
        raise
    os.remove(tmpfile)


def GitFile(filename, mode='rb', bufsize=-1):
    """Create a file object that obeys the git file locking protocol.

    :return: a builtin file object or a _GitFile object

    :note: See _GitFile for a description of the file locking protocol.

    Only read-only and write-only (binary) modes are supported; r+, w+, and a
    are not.  To read and write from the same file, you can take advantage of
    the fact that opening a file for write does not actually open the file you
    request.
    """
    if 'a' in mode:
        raise IOError('append mode not supported for Git files')
    if '+' in mode:
        raise IOError('read/write mode not supported for Git files')
    if 'b' not in mode:
        raise IOError('text mode not supported for Git files')
    if 'w' in mode:
        return _GitFile(filename, mode, bufsize)
    else:
        return io.open(filename, mode, bufsize)


class _GitFile(object):
    """File that follows the git locking protocol for writes.

    All writes to a file foo will be written into foo.lock in the same
    directory, and the lockfile will be renamed to overwrite the original file
    on close.

    :note: You *must* call close() or abort() on a _GitFile for the lock to be
        released. Typically this will happen in a finally block.
    """

    PROXY_PROPERTIES = set(['closed', 'encoding', 'errors', 'mode', 'name',
                            'newlines', 'softspace'])
    PROXY_METHODS = ('__iter__', 'flush', 'fileno', 'isatty', 'read',
                     'readline', 'readlines', 'seek', 'tell',
                     'truncate', 'write', 'writelines')

    def __init__(self, filename, mode, bufsize):
        self._filename = filename
        self._lockfilename = '%s.lock' % self._filename
        retry = True
        while retry:
            try:
                fd = os.open(
                    self._lockfilename,
                    os.O_RDWR | os.O_CREAT | os.O_EXCL | getattr(os, "O_BINARY", 0))

                retry = False
            except FileExistsError:
                time.sleep(0.1)
                os.remove(self._lockfilename)

        self._file = os.fdopen(fd, mode, bufsize)
        self._closed = False

        for method in self.PROXY_METHODS:
            setattr(self, method, getattr(self._file, method))

    def abort(self):
        """Close and discard the lockfile without overwriting the target.

        If the file is already closed, this is a no-op.
        """
        if self._closed:
            return
        self._file.close()
        try:
            os.remove(self._lockfilename)
            self._closed = True
        except OSError as e:
            # The file may have been removed already, which is ok.
            if e.errno != errno.ENOENT:
                raise
            self._closed = True

    def close(self):
        """Close this file, saving the lockfile over the original.

        :note: If this method fails, it will attempt to delete the lockfile.
            However, it is not guaranteed to do so (e.g. if a filesystem
            becomes suddenly read-only), which will prevent future writes to
            this file until the lockfile is removed manually.
        :raises OSError: if the original file could not be overwritten. The
            lock file is still closed, so further attempts to write to the same
            file object will raise ValueError.
        """
        if self._closed:
            return
        self._file.close()
        try:
            try:
                os.rename(self._lockfilename, self._filename)
            except OSError as e:
                if sys.platform == 'win32' and e.errno == errno.EEXIST:
                    # Windows versions prior to Vista don't support atomic
                    # renames
                    _fancy_rename(self._lockfilename, self._filename)
                else:
                    raise
        finally:
            self.abort()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __getattr__(self, name):
        """Proxy property calls to the underlying file."""
        if name in self.PROXY_PROPERTIES:
            return getattr(self._file, name)
        raise AttributeError(name)



# -*- coding: utf8 -*-
import datetime


class UpdateTimer(object):
    def __init__(self, interval=1.):
        self.__interval = interval
        self.__start_time = None

    def can_update(self):
        now = datetime.datetime.utcnow()
        if self.__start_time is None or (now - self.__start_time).total_seconds() > self.__interval:
            self.__start_time = now
            return True

        return False



# -*- coding: utf8 -*-
import json
import logging
import os
from .path_utils import has_moniker


class MetadataFiles(object):
    metadata_ext = '.metadata.json'

    @classmethod
    def __meta_companion_file_name(cls, metadata_file_name):
        return metadata_file_name[:-len(cls.metadata_ext)]

    @classmethod
    def __is_meta_folder_file_name(cls, metadata_file_name):
        return os.path.basename(metadata_file_name).lower() == 'folder' + cls.metadata_ext

    @classmethod
    def __convert_data_unsupported_type(cls, data):
        for key, val in data.items():
            if isinstance(val, (dict, list)):  # we convert arrays and dicts into string as we don't support them yet
                data[key] = json.dumps(val)

    @classmethod
    def __handle_json_file(cls, metadata_json_file, on_data):
        from .data_sync import InvalidJsonFile

        with open(metadata_json_file) as metadata_file:
            try:
                data = json.load(metadata_file)
            except ValueError as ex:
                raise InvalidJsonFile(metadata_json_file, ex)

            return on_data(data)

    @classmethod
    def __get_metadata_info(cls, repo, rel_metadata_file_name):
        full_path_metadata = os.path.join(repo.data_path, rel_metadata_file_name)

        if has_moniker(full_path_metadata):
            return rel_metadata_file_name, {}

        rel_path_key = rel_metadata_file_name[:-len(cls.metadata_ext)]

        def handle_data(data):
            cls.__convert_data_unsupported_type(data)

            return rel_path_key, data

        return cls.__handle_json_file(full_path_metadata, handle_data)

    @classmethod
    def __get_folder_metadata_info(cls, repo, rel_metadata_file_name):
        full_path_metadata = os.path.join(repo.data_path, rel_metadata_file_name)

        rel_path = os.path.relpath(os.path.dirname(full_path_metadata), repo.data_path)

        if rel_path == '.':
            rel_path = ''

        def handle_data(all_files_data):
            for filename, file_data in all_files_data.items():
                cls.__convert_data_unsupported_type(file_data)
                rel_path_key = os.path.join(rel_path, filename)

                yield rel_path_key, file_data

        return cls.__handle_json_file(full_path_metadata, handle_data)

    @classmethod
    def __get_metadata(cls, repo, data_files_info, metadata_rel_file_name):
        if cls.__is_meta_folder_file_name(metadata_rel_file_name):
            for rel_path_key, data in cls.__get_folder_metadata_info(repo, metadata_rel_file_name):
                yield rel_path_key, data
        elif cls.__meta_companion_file_name(metadata_rel_file_name) in data_files_info:
            rel_path_key, data = cls.__get_metadata_info(repo, metadata_rel_file_name)

            yield rel_path_key, data
        else:
            logging.debug("file %s doesn't have a data point file", metadata_rel_file_name)

    @classmethod
    def load_all_metadata(cls, repo, data_files_info, metadata_files_list, no_progressbar=False):
        from tqdm import tqdm

        files_metadata = {}
        if len(metadata_files_list) == 0:
            return files_metadata

        for metadata_rel_file_name in tqdm(metadata_files_list, desc='Read metadata', unit=' files', ncols=80, disable=no_progressbar):
            for rel_path_key, data in cls.__get_metadata(repo, data_files_info, metadata_rel_file_name):
                files_metadata.setdefault(rel_path_key, {}).update(data)

        return files_metadata



import hashlib
import logging
import re
import tempfile
import uuid
from os import path, remove
from shutil import copyfile
from traceback import format_exception_only

from six.moves.urllib import parse
from mali_commands.rsync import makedir
from shutil import rmtree, copytree


def export_exception(ex):
    return ('\n'.join(format_exception_only(ex.__class__, ex))).strip()


class GitError(Exception):
    pass


class GitExpando(object):
    pass


def _append_commit_link(remote_template):
    x = parse.urlparse(remote_template)
    if x.hostname is not None:
        remote_template += '/commit/{}' if x.hostname.lower() == 'github.com' else '/commits/{}'
    return remote_template


def ssh_to_https(url):
    if '@' in url:
        return 'https://' + (url.split('@')[1].strip()).replace(':', '/')
    return url


def enforce_https(url):
    if url.startswith('http://'):
        return 'https://' + url[len('http://'):]
    return url


def remove_dot_git(url):
    return url if not url.lower().endswith('.git') else url[:-4]


def remote_to_template(remote_url):
    return remove_dot_git(enforce_https(ssh_to_https(remote_url.lower())))


def get_remote_url():
    try:
        import git
        repo = git.Repo('.', search_parent_directories=True)
        if len(repo.remotes) > 0:
            remote = list(repo.remotes)[0]
            return remote_to_template(remote.url)
    except Exception as gr:
        logging.error("Failed to get remote git repository of current folder. {}".format(str(gr)))
    return None


def remote_url_obj(url):
    if url is None:
        return None

    obj = parse.urlparse(remote_to_template(url))
    if obj.hostname is None or obj.hostname == '':
        return None

    return obj


def get_org_from_remote_url(url):
    url_obj = remote_url_obj(url)
    if url_obj is None:
        return None

    return '/'.join(url_obj.path.split('/')[:-1]).lower()


def validate_diff_urls(src, trg):
    src = remote_to_template(src)
    trg = remote_to_template(trg)
    return src == trg


def validate_remote_urls_org(src, trg):
    src_org = get_org_from_remote_url(src)
    remote_org = get_org_from_remote_url(trg)

    return src_org == remote_org


def __fill_repo(repo, res=GitExpando()):
    res.repo = repo
    res.git_version = repo.git.version()
    res.remote_name = None
    res.remote_url = None
    res.has_head = repo.head.is_valid()
    res._remote_template = None
    if len(repo.remotes) > 0:
        remote = list(repo.remotes)[0]
        res.remote_name = remote.name
        res.remote_url = remote.url
        res._remote_template = _append_commit_link(remote_to_template(remote.url))
    res.branch = repo.active_branch
    res.head_sha = res.repo.head.object.hexsha if res.has_head else None

    res.head_sha_url = res._remote_template.format(res.head_sha) if res._remote_template else res.head_sha
    res.is_clean = repo.git.status(porcelain=True).strip() == ''
    return res


def _export_commit(commit):
    return {'author': commit.author.name,
            'email': commit.author.email,
            'date': commit.authored_datetime.isoformat()}


def import_git():
    try:
        import git
        return git
    except ImportError:
        logging.warning('Failed to import git')
    raise GitError('Failed to import git')


def get_repo(path_=None, repo=None):
    git = import_git()
    try:
        repo = repo or git.Repo(path_, search_parent_directories=True)
        response = __fill_repo(repo)
        response.export_commit = _export_commit
        response.refresh = lambda: __fill_repo(repo, response)
        return response
    except git.exc.GitCommandError as ex:
        logging.warning('Failed to get repo status: {}'.format(export_exception(ex)))
        raise GitError('Failed to get repo status: {}'.format(export_exception(ex)))
    except git.exc.InvalidGitRepositoryError:
        logging.warning('path {} is not tracked'.format(path_))
        raise GitError('path is not tracked')


def try_get_raw_repo(path_='.'):
    try:
        repo = import_git().Repo(path_, search_parent_directories=True)
        return remote_to_template(list(repo.remotes)[0].url)
    except GitError as ex:
        logging.debug('Failed to get local repo path. got: {}'.format(export_exception(ex)))
    except Exception as ex:
        logging.warning('Failed to get local repo path. got: {}'.format(export_exception(ex)))
    return None


class GitRepoSyncer(object):
    @classmethod
    def _sanitize_branch_name(cls, name):
        return re.sub(r'[^a-zA-Z0-9]', "_", name)

    @classmethod
    def try_ex_path(cls, git, repo_path):
        if path.isdir(repo_path):
            try:
                repo = git.Repo(repo_path)
                return repo
            except Exception as ex:
                logging.warning("Failed to init tracking repository repo at {}, got {}".format(repo_path, str(ex)))
                remove(repo_path)
        return None

    @classmethod
    def _validate_tracking_repository_or_clone(cls, repo_path, tracking_origin_url):

        git = import_git()
        repo = cls.try_ex_path(git, repo_path) or git.Repo.clone_from(tracking_origin_url, repo_path)
        for remote in repo.remotes:
            if remote.name == 'origin' and remote.url != tracking_origin_url:
                raise GitError("Tracking repository at {} has remote other than {} tracking_origin_url configured: {}".format(repo_path, tracking_origin_url, remote.url))

        repo.git.reset(hard=True)
        return repo

    @classmethod
    def clone_tracking_repo(cls, tracking_repository_remote_url, fixed_clone_path=None):
        repo_id = 'missinglink_{}'.format(hashlib.sha256(tracking_repository_remote_url.encode('utf-8')).hexdigest())
        temp_path = fixed_clone_path or path.join(tempfile.gettempdir(), repo_id)
        logging.debug("Clone {} to {}".format(tracking_repository_remote_url, temp_path))
        return cls._validate_tracking_repository_or_clone(temp_path, tracking_repository_remote_url)

    @classmethod
    def _checkout(cls, repo, branch=None):
        if branch in repo.branches:
            repo.git.checkout(branch)
        else:
            repo.git.checkout(b=branch)

    @classmethod
    def _get_changed_files(cls, repo):
        changes = repo.git.status(porcelain=1).strip()
        if len(changes) == 0:
            return []
        files = list(map(lambda x: x.split(' ')[-1].strip(), repo.git.status(porcelain=1).strip().split('\n')))
        return files

    @classmethod
    def _copy_path(cls, src_file, tracked_file):
        if not path.exists(src_file):
            return False

        if path.isdir(src_file):
            if path.exists(tracked_file):
                rmtree(tracked_file, ignore_errors=True)

            copytree(src_file, tracked_file)
            return True

        copyfile(src_file, tracked_file)
        return True

    @classmethod
    def _remove_path(cls, tracked_file):
        if not path.exists(tracked_file):  # the file is deleted - delete it
            return

        if path.isdir(tracked_file):
            rmtree(tracked_file)
            return

        remove(tracked_file)
        return

    @classmethod
    def _sync_uncommitted_changes(cls, src, tracking):
        changed_files = cls._get_changed_files(src)
        for file in changed_files:
            src_file = path.join(src.working_dir, file)
            tracked_file = path.join(tracking.working_dir, file)
            logging.debug('_sync_uncommitted_changes %s -> %s', src_file, tracked_file)

            if not cls._copy_path(src_file, tracked_file):  # the file is present, copy it...
                cls._remove_path(tracked_file)

        if len(changed_files) > 0:
            tracking.git.add('.')
            tracking.index.commit('ML AI: Synced file(s) from {}.Files: \n{}'.format(src.working_dir, '\n'.join(changed_files)))

    @classmethod
    def _merge_to_new_branch(cls, src, tracking, branch_name):

        cls._checkout(tracking, 'dev_null')
        tracking.git.fetch('origin')
        tracking.git.fetch('src')
        cls._checkout(tracking, branch_name)
        # OK, so why not to sync the current src branch to the target one and then move to the custom branch?
        # Because of branch name collisions
        # lets say Alice and Bob are pullig the src from the same head `HEAD1` (rev: `REV0`) and checkout new branch `BR1`
        # in `BR1` they both create a `filename.txt` , write their name into the file and commit (but not push) (commit ids `A1...` and `B1...`)
        # then they both run repo sync and we attempt to sync
        # When Alice pushes to the tracking repo, the `BR1` label will point to her commit
        # When Bob tries to run sync, he will get her branch and his commit will have to either override her (problematic to implement and unreliable)
        # or be atop of her commit (the shadow repo wont reflect the truth)
        # In conclusion, the "correct" way to do it is to create the internal branch (`T1` and `T2`)  and then "merge" the src branch into it
        # this will copy the shared commit tree from the src repo (upto and including `REV0`) and as it is pushed to the original repo, it will have the same order on both (even without the branch names)
        # due to the fact that commits are in an ordered tree
        # once the history upto `REV0` is merged (the commits ids are still the same) the un-pushed commit ids live only in the `T*` branches and on top of them
        # we are adding the sync commit
        # As we never merge between shadow branches the commit order of the src repository nor its Ids are ever changed
        # and the commit log of every shadow branch is synced_commit->un-pushed commits (to src)->pushed commits (to the src)
        if src.head.is_valid():
            tracking.git.merge('src/{}'.format(src.active_branch), allow_unrelated_histories=True)
            return True
        return False

    @classmethod
    def _push_to_tracking_repository_remote(cls, tracking_repository):
        if not tracking_repository.head.is_valid():
            tracking_repository.git.commit(m='empty commit', allow_empty=True)
        tracking_repository.remotes['origin'].push('{}:{}'.format(tracking_repository.active_branch, tracking_repository.active_branch))

    @classmethod
    def merge_src_to_tracking_repository(cls, src, tracking_repository, br_tag=None):
        git = import_git()
        if git.Remote(src, 'src') not in tracking_repository.remotes:
            tracking_repository.create_remote('src', src.git_dir)
        br_tag = br_tag or uuid.uuid4().hex[0:5]

        br_name = '{}_{}'.format(br_tag, src.active_branch)
        br_name = 'invocation/' + cls._sanitize_branch_name(br_name)
        cls._merge_to_new_branch(src, tracking_repository, br_name)
        cls._sync_uncommitted_changes(src, tracking_repository)
        cls._push_to_tracking_repository_remote(tracking_repository)
        cls._checkout(tracking_repository, 'dev_null')
        return br_name
